/* 
 * Keykerictl - Keykeriki control tool
 * Dump terminal
 * 
 * Thorsten Schroeder <ths  (at)   dreamlab.net>
 * 20090512, Berlin, Germany
 *
 * BeerWare license.
 *
 * Revision: $Rev: 35 $
 * FILEID: $Id: keykerictl.c 35 2009-06-07 20:19:57Z ths $
 *
 * build: gcc -Wall -ansi -D__USE_BSD -o keykerikictl keykerikictl.c 
 * Tested environments:
 *
 *    + 20090512  - ths    - Ubuntu Linux 9.04     - OK 
 *    + 20090512  - ths    - FreeBSD 7.2 STABLE    - Not yet OK, Work in
 *                                                   progress
 *
 * */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/time.h>

#define CANCEL 'X'-64
#define ERROR  -1

int               serial_fd;
struct termios    ttyio, kbdios;

const char _file_rev[] = "$Rev: 35 $";

void cb_onexit(void);
void kbd_setraw(void);
int serial_open(char *, int);

int serial_open(char *device, int baudrate)
{
   int baudflag;
   int ret = 0;

   serial_fd=open(device, O_RDWR | O_NDELAY);

   if(serial_fd==ERROR) {
      printf("Unable to open %s\n", device);
      return(ERROR);
   }
   /* set attributes of serial interface to raw mode */
   if(tcgetattr(serial_fd, &ttyio)==ERROR) {
      printf("tcgetattr() error (%d)\n", __LINE__);
      close(serial_fd);
      return(ERROR);
   }
   ttyio.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
   ttyio.c_oflag &= ~OPOST;

   /* if the following line is commented out, serial input is line buffered on
    * FreeBserial_fd. */
   ttyio.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);

   ttyio.c_cflag &= ~(CSIZE|PARENB);
   ttyio.c_cflag |= CS8;
   
   /* set baudrate */
   switch(baudrate) {
      case 115200: 
         baudflag=B115200; 
         break;
      case 57600: 
         baudflag=B57600; 
         break;
      case 38400: 
         baudflag=B38400; 
         break;
      case 19200: 
         baudflag=B19200; 
         break;
      case 4800: 
         baudflag=B4800; 
         break;
      default: 
         baudrate=9600; 
         baudflag=B9600; 
         break;
   }

   cfsetispeed(&ttyio, baudflag);
   cfsetospeed(&ttyio, baudflag);

   if(tcsetattr(serial_fd, TCSANOW, &ttyio)==ERROR) {
      printf("tcsetattr() error (%d)\n", __LINE__);
      close(serial_fd);
      return(ERROR);
   }

   fprintf(stderr, "[*] Keykerictl - %s\n", _file_rev);
   fprintf(stderr, "[*] Using device %s @ %d baud\n", device, baudrate);
   fprintf(stderr, "[*] Connected to Keykeriki\n");
   fprintf(stderr, "[*] Press Ctrl-X to exit\n");

   ret=write(serial_fd, "\r\nhelp\r\n", 8);
   return(0);
}

int main(int argc, char **argv)
{
   char data[4], *device;
   int n;
   int baudrate;
   fd_set fdset;
   int ret = 0;

   if(!isatty(STDIN_FILENO)) {
     fprintf(stderr, "[!] input must be from a keyboard.\n");
     return(ERROR);
   }

   if(argc < 3) {
      fprintf(stderr, "[!] usage: %s <serial device> <baudrate>\n", argv[0]);
      exit(ERROR);
   }

   /* open serial port */

   baudrate = atoi(argv[2]);
   device   = argv[1];

   if(serial_open(device, baudrate)) 
      return(1);
   
   kbd_setraw();

   /* Now we may use the serial interface and keyboard for sending
     just one character at a time */


   do {
      /* Waiting for data on serial interface (serial_fd) or keyboard (STDIN_FILENO) */
      FD_ZERO(&fdset);
      FD_SET(serial_fd,  &fdset);
      FD_SET(STDIN_FILENO, &fdset);

      select(FD_SETSIZE, &fdset, NULL, NULL, NULL);

      while((n=read(serial_fd, data, 1)) > 0) {

         ret=write(STDIN_FILENO, data, n);
         if(ret<0) {
            fprintf(stderr, "[!] write error (line %d).\n", __LINE__);
            break;
         }
         /*printf("%c", *data); <- XXX wenn man das macht, gibt er auf FreeBserial_fd nur aus,
           wenn ein newline kam!! write(STDIN_FILENO, data, n); machts
           direkt */
      }
      
      *data = '\0';
      
      /* XXX das funktioniert auf luniks/ubuntu jedoch nicht auf FreeBserial_fd */
      while((n=read(STDIN_FILENO, data,4)) > 0) { 
         ret=write(serial_fd, data, n);
         if(ret<0) {
            fprintf(stderr, "[!] write error (line %d): %d (%s)\n", 
                  __LINE__, errno, strerror(errno));
            break;
         }
      }

   } while(*data != CANCEL);
   
   close(serial_fd);
   return(0);
}

/* exit callback */
void cb_onexit(void)
{
   /* restore TTYP mode on exit */
   /*fcntl(STDIN_FILENO, 0, FNDELAY);*/
   fcntl(STDIN_FILENO, 0, O_NDELAY);
   tcsetattr(STDIN_FILENO, TCSANOW, &kbdios);
   printf("[*] KTHXBYE\n");
}

void kbd_setraw(void)
{
   tcgetattr(STDIN_FILENO, &kbdios);
   atexit(cb_onexit);
   memcpy(&ttyio, &kbdios, sizeof(struct termios));
   ttyio.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
   tcsetattr(STDIN_FILENO, TCSANOW, &ttyio);

   /* read shall return ASAP, even if there is no data to be read. */
   /*fcntl(serial_fd, F_SETFL, FNDELAY);*/
   fcntl(serial_fd, F_SETFL, O_NDELAY);
   /*fcntl(STDIN_FILENO, F_SETFL, FNDELAY);*/
   fcntl(STDIN_FILENO, F_SETFL, O_NDELAY);
}

