/* 
 * $Id: sd.c 32 2009-06-07 20:11:05Z ths $
 *
 * SD Card access
 *
 * Author(s)/Copyright:
 *    Thorsten Schroeder <ths  at  dev.io>
 *
 * Revision: $Rev: 32 $
 *
 * Germany Berlin, 2009/04/24
 *
 * */
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <string.h>
#include <avr/eeprom.h>
#include "sd.h"
#include "eeprom.h"
#include "main.h"

extern uint8_t buf[];
extern uint8_t buf_idx;       /* current index in buffer buf */
extern unsigned long sd_addr; /* SD card sector address/byte offset */

extern keykeriki_config config;

static void    sd_byte_write(uint8_t);
static uint8_t sd_sector_write(unsigned long, uint8_t *);
static uint8_t sd_sector_read(unsigned long, uint8_t *);
static uint8_t sd_byte_read(void);
static uint8_t sd_cmd_write(uint8_t *);
static uint8_t sd_flush(void);
static uint16_t sd_block_read(uint8_t *, uint8_t *, uint16_t);

#if 0
static uint8_t sd_csd_read(uint8_t *);
#endif

#define MAX_TRIES 200   /* SD card initialization */

uint8_t sd_init(void)
{
   uint16_t it = 0;
   uint8_t ret = 0;

   printf_P(PSTR("\r\n[*] SD memory card init: "));

   do {
      uint8_t i = 0;
      uint16_t init_tries = 0;
      uint8_t cmd[] = {0x40,0x00,0x00,0x00,0x00,0x95};

      memset(&buf, 0, DATA_BLOCK_SIZE);
      buf_idx=0;

      SD_DDR &= ~(1<<SD_MISO);
      SD_DDR |=  (1<<SD_MOSI);
      SD_DDR |=  (1<<SD_SCK); 
      SD_DDR |=  (1<<SD_CS); 
      SD_PORT |= (1<<SD_CS);

      _delay_ms(200);

      /* SPI enable */
      SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0)|(1<<SPR1);
      SPSR = (0<<SPI2X);

      for(i=0; i<16; i++)
         sd_byte_write(0xFF);
      
      /* cmd 0 */
      while( sd_cmd_write(cmd) != 1 )
         if( init_tries++ > MAX_TRIES ) {
            SD_DISABLE();
            ret=1;
            break;
         }


      init_tries=0;

      /* cmd 1 */
      cmd[0] = 0x41;
      cmd[5] = 0xff;

      while( sd_cmd_write(cmd) != 0 )
         if( init_tries++ > (MAX_TRIES * 2) ) {
            SD_DISABLE();
            ret=2;
            break;
         }


      /* SPI speed */
      SPCR &= ~((1<<SPR0) | (1<<SPR1));
      SPSR = SPSR|(1<<SPI2X);

      printf_P(PSTR("%02x "), ret);

   } while(ret != 0 && it++ < 16); 

   SD_DISABLE();

   if(ret == 0) {
      printf_P(PSTR("... done.\r\n"));
      if(config.sd_append) {
         sd_addr = sd_last_sector()/DATA_BLOCK_SIZE;
      } else
         sd_addr = 1;
   } else {
      printf_P(PSTR("... failed.\r\n"));
      sd_disable(0);
   }

   return(ret);
}


unsigned long sd_last_sector(void)
{
   unsigned long addr = 0;
   uint8_t tmp[DATA_BLOCK_SIZE];

   /* read fs header (first sector at addr 0x0) */
   sd_sector_read(0, tmp);

   memcpy(&addr, tmp, 4);

   return(addr/DATA_BLOCK_SIZE);
}

uint8_t sd_store(uint8_t *data, uint16_t len)
{

   uint16_t i=0;
   uint8_t ret=0;
   
   uint8_t *pbuf;

   pbuf=buf;
   pbuf+=(buf_idx<<4); /* x RECORD_SIZE */

   if(len > RECORD_SIZE)
      len=RECORD_SIZE;

   for(i=0; i<len; i++) {
      *pbuf++ = *data++;
   }

   buf_idx++;

   if(buf_idx >= (DATA_BLOCK_SIZE>>4)) {/* max records */
      ret=sd_flush();
   }

   return(ret);
}

static uint8_t sd_flush(void)
{
   uint8_t ret = 0;

   ret = sd_sector_write(sd_addr, buf);

   if(buf_idx >= (DATA_BLOCK_SIZE>>4)) {/* max records */
      sd_addr++;
      memset(&buf, 0, DATA_BLOCK_SIZE);
      buf_idx=0;
   }   

   return(ret);
}

uint8_t sd_fs_update_header(unsigned long addr, uint8_t *data)
{

   uint8_t cmd[] = CMD24;
   uint8_t saddr[4];
   uint8_t ret = 0;
   uint8_t i = 0;
   uint16_t j = 0;

   addr = addr*DATA_BLOCK_SIZE;

   cmd[1] = 0;
   cmd[2] = 0;
   cmd[3] = 0;

   saddr[3] = ((addr & 0xFF000000) >>24 );
   saddr[2] = ((addr & 0x00FF0000) >>16 );   
   saddr[1] = ((addr & 0x0000FF00) >>8 );    
   saddr[0] = ((addr & 0x000000FF));    


   ret = sd_cmd_write(cmd);
   
   /* err */
   if(ret != 0)
      return(ret);

   /* pause.. */
   for(i=0;i<100;i++)
      sd_byte_read();

   /* start byte */
   sd_byte_write(0xFE);

   /* write block */
   for(j=0; j<sizeof(saddr);j++) {
      sd_byte_write(saddr[j]);
   }

   /* write block */
   for(j=0; j<(DATA_BLOCK_SIZE>>1);j++) {
      sd_byte_write(0x20);
      sd_byte_write(0x2c);
      /* abschaffen. */
   }

   /* CRC not used */
   sd_byte_write(0xFF);
   sd_byte_write(0xFF);

   /* err */
   if( (sd_byte_read() & 0x1F) != 5 )
      return(1);

   /* wait for sd card */
   while(sd_byte_read() != 0xFF);

   SD_DISABLE();

   return(0);
}

static uint8_t sd_sector_read(unsigned long addr, uint8_t *data)
{
   uint8_t cmd[] = {0x51,0x00,0x00,0x00,0x00,0xFF};

   addr = addr*DATA_BLOCK_SIZE;

   cmd[1] = ((addr & 0xFF000000)>>24);
   cmd[2] = ((addr & 0x00FF0000)>>16);
   cmd[3] = ((addr & 0x0000FF00)>>8);

   sd_block_read(cmd, data, DATA_BLOCK_SIZE);
   return(0);
}


static uint8_t sd_sector_write(unsigned long addr, uint8_t *data)
{
   uint8_t cmd[] = CMD24;
   uint8_t ret = 0;
   uint8_t i = 0;
   uint16_t j = 0;

   addr = addr*DATA_BLOCK_SIZE;

   cmd[1] = ((addr & 0xFF000000) >>24 );
   cmd[2] = ((addr & 0x00FF0000) >>16 );
   cmd[3] = ((addr & 0x0000FF00) >>8 );

   ret = sd_cmd_write(cmd);
   
   /* err */
   if(ret != 0) {
      return(ret);
   }

   /* pause.. */
   for(i=0;i<100;i++)
      sd_byte_read();

   /* start byte */
   sd_byte_write(0xFE);

   /* write block */
   for(j=0; j<DATA_BLOCK_SIZE;j++)
      sd_byte_write(*data++);

   /* CRC not used */
   sd_byte_write(0xFF);
   sd_byte_write(0xFF);

   /* err */
   if( (sd_byte_read() & 0x1F) != 5 )
      return(1);

   /* wait for sd card */
   while(sd_byte_read() != 0xFF);

   SD_DISABLE();

   sd_fs_update_header(addr, NULL);

   return(0);
}

static uint8_t sd_cmd_write(uint8_t *cmd)
{
   uint16_t tries = 0;
   uint8_t ret = 0xff;
   uint8_t i = 0;

   SD_DISABLE();
   
   /* send 8 clk */
   sd_byte_write(0xFF);
   
   SD_ENABLE();

   for(i=0; i<6; i++) {
      sd_byte_write(*cmd++);
   }

   while(ret == 0xff) {
      ret = sd_byte_read();
      if(tries++ > 500)
         break;
   }
   return(ret);
}


static void sd_byte_write(uint8_t b)
{
   SPDR=b;

   /* wait SPI/SD */
   while(!(SPSR & (1<<SPIF)));
}


static uint8_t sd_byte_read(void)
{
   uint8_t ret = 0;
   
   SPDR = 0xff;

   /* wait SPI/SD */
   while(!(SPSR & (1<<SPIF)));

   ret = SPDR;
   return(ret);
}

#if 0
static uint8_t sd_csd_read(uint8_t *buf)
{
   uint8_t ret = 0;
   uint8_t cmd[] = {0x49,0x00,0x00,0x00,0x00,0xFF};

   ret=sd_block_read(cmd, buf, 16);
   return(ret);
}
#endif

static uint16_t sd_block_read(uint8_t *cmd, uint8_t *buf, uint16_t len) 
{
   uint8_t ret = 0;
   uint16_t i = 0;

   ret=sd_cmd_write(cmd);
   if(ret != 0)
      return(ret);

   /* wait for start byte */
   while(sd_byte_read() != 0xfe);

   for(i=0;i<len;i++) {
      buf[i] = sd_byte_read();
   }

   /* ignore CRC */
   sd_byte_read();
   sd_byte_read();

   SD_DISABLE();
   return(i);
}

void sd_disable(uint8_t p)
{
   config.enable_sd= 0;
   if(p) {
      eeprom_busy_wait();
      eeprom_write_byte(&ee_cfg_enable_sdcard, config.enable_sd);
   }
}

void sd_enable(uint8_t p)
{
   config.enable_sd= 1;
   if(p) {
      eeprom_busy_wait();
      eeprom_write_byte(&ee_cfg_enable_sdcard, config.enable_sd);
   }
}

uint8_t sd_append(void)
{
   config.sd_append ^= 1;
   eeprom_busy_wait();
   eeprom_write_byte(&ee_cfg_enable_sdcard_append, config.sd_append);
   return(config.sd_append);
}

