
/* 
 * $Id: twi_master.c 32 2009-06-07 20:11:05Z ths $
 *
 * Parts of this code comes from a "roboter forum" without any 
 * copyright/author header/information.
 *
 * Thorsten Schroeder <ths  (at)  dev.io>
 * Germany, Berlin, 2009/04/22
 *
 * $Rev: 32 $
 *
 * */

#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include <stdio.h>
#include <util/delay.h>

#ifndef _TWI_MASTER
#define _TWI_MASTER
#endif

#include "twi.h"

#define FREQ_ 100000

uint8_t twi_master_init(uint32_t scl_freq)
{
	/* set bitrate - http://www.roboternetz.de/wissen/index.php/TWI#Bit_Rate_Generator */
	
	TWBR = (( (F_CPU/FREQ_) - 16) / 2);
	twi_master_activate_pullups();
	
   printf_P(PSTR("[D] twi init: TWBR = %x\r\n"),  TWBR);
	return(TWBR > 10); /* TRUE/FALSE -> success/failed */
}

/* activate internal pull-up resistors for SCL/SDA. Otherwise external pull-up resistors must be applied. */
void twi_master_activate_pullups(void)
{
	/* activate internal pull-ups for SCL/SDA */
	TWI_PORT |= ((1<<TWI_SCL)|(1<<TWI_SDA));
	
	return;
}

uint8_t twi_master_start(uint8_t addr, uint8_t op)
{
	uint8_t status = 0x00;
	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
	
	/* Wait until completed */
	while( !(TWCR & (1<<TWINT)) );
	
	/* get the first 5 bits (result) from status register (TWSR) */
	status = TWSR & 0xF8;
	
	if( (status != TWI_START) && (status != TWI_REP_START))
		return(0);
	
	/* send dest. addr  */
	/* TWDR -> Data Register	*/
	TWDR = (addr<<1) | op;
	TWCR = (1<<TWINT)|(1<<TWEN);
	
	/* wait for ACK/NACK */
	while (!(TWCR & (1<<TWINT)));
	
	/* get status bits 11111000b */
	status = TWSR & 0xF8;
	
	if((status != TWI_MTX_ADR_ACK) && (status != TWI_MRX_ADR_ACK))
		return(0);
		
	return(1);	
}

/* stop master */
void twi_master_stop(void)
{
	/* send stop command TWSTO */
	TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
	
	/* wait for success */
	while (TWCR & (1<<TWINT));
	
}

/* send one byte to the slave */
uint8_t twi_master_write_byte(uint8_t b)
{
	uint8_t status = 0;
	
	/* set byte in data register & send */
	TWDR = b;
	TWCR = (1<<TWINT)|(1<<TWEN);
	
	/* wait for success */
	while (!(TWCR & (1<<TWINT)));
	
	status = TWSR & 0xF8;
	if (status != TWI_MTX_DATA_ACK) 
		return 0;
	
	return(1);
}


/* params: address, buffer to write, size of buffer 
 max 255 bytes
*/
uint8_t twi_master_write(uint8_t slave, unsigned char *buf, uint8_t size)
{
	uint8_t i = 0;
	uint8_t status = 0;
	
	status = twi_master_start(slave, TWI_WRITE);

	if(status == 0){
		twi_master_stop();
		printf_P(PSTR("\r\nMift... WRITE hatnichtgeklappt...\r\n"));
	} else {
		for(i=0; i<size && i<0xff; i++) {
			twi_master_write_byte(buf[i]);
		}
		twi_master_stop();
	}
	return(i+1);
}


/* read ACK byte from twi */
uint8_t twi_master_read_ack(void)
{
	TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWEA);
	while (!(TWCR & (1<<TWINT)));
	
	/* return the read byte which is in the data register now */
	return(TWDR);
		
}

/* read NACK(??) byte from twi */
uint8_t twi_master_read_nack(void)
{
	TWCR = (1<<TWINT)|(1<<TWEN);
	while(!(TWCR & (1<<TWINT)));
	
	/* return the read byte which is in the data register now */
	return(TWDR);
		
}


/* read size bytes and store it in buffer buf which must be 
 able to hold the data.
 returns count
 */
uint8_t twi_master_read(uint8_t slave, unsigned char *buf, uint8_t maxsize)
{
	uint8_t i = 0;
	uint8_t status = 0;

	/* 1. read something  */
	status = twi_master_start(slave, TWI_READ);

	if(status == 0){
		twi_master_stop();
		printf_P(PSTR("\r\nMift... READ hatnichtgeklappt...\r\n"));
	} else {
		for(i=1; (i<maxsize); i++) {
			buf[i-1] = twi_master_read_ack();
		}
		buf[i-1] = twi_master_read_nack();
		twi_master_stop();
	}
	return(i+1);
}

